from __future__ import annotations
import os.path
import sys
from typing import Dict, List, Optional, Tuple, Union

from git import Commit, Diff, DiffIndex, Repo

from shulkr.java import get_renamed_variables


def get_blob(commit: Optional[Commit], path: str, repo_path: str) -> bytes:
	if commit is None:
		p = os.path.join(repo_path, path)
		with open(p, 'r') as f:
			return f.read()

	parts = path.split('/')
	curr = commit.tree

	while len(parts) > 0:
		name = parts.pop(0)
		curr = curr[name]

	return curr.data_stream.read().decode()



import os
import sys

from git import InvalidGitRepositoryError, Repo

from shulkr import parse_args, create_version
from shulkr.minecraft.version import NoSuchVersionError, Version, load_manifest


def main_uncaught() -> None:
	load_manifest()

	args = parse_args()

	repo_path = os.path.join(
		os.getcwd(),
		args.repo
	)

	try:
		ranges = [Version.pattern(p) for p in args.version]
		versions = [version for range in ranges for version in range]
		versions.sort()
	except NoSuchVersionError as e:
		print(e, file=sys.stderr)
		sys.exit(1)

	if not os.path.exists(repo_path):
		print(f'Creating {repo_path}')
		os.mkdir(repo_path)

	try:
		repo = Repo(repo_path)
	except InvalidGitRepositoryError:
		print('Initializing git')
		repo = Repo.init(repo_path)

	for version_id in versions:
		create_version(
			repo,
			version_id,
			args.undo_renamed_vars,
			args.message
		)


def main() -> None:
	try:
		main_uncaught()
	except KeyboardInterrupt:
		print('Aborted!', file=sys.stderr)


if __name__ == '__main__':
	main()



import argparse
import os.path
import sys

from git import BadName, InvalidGitRepositoryError, Repo

from shulkr.git import get_blob
from shulkr.java import JavaAnalyzationError, get_renamed_variables, undo_variable_renames
from shulkr.minecraft.source import generate_sources
from shulkr.minecraft.version import Version


def parse_args() -> argparse.Namespace:
	parser = argparse.ArgumentParser(prog='shulkr', description='Generate multiple versions of the Minecraft source code')
	parser.add_argument('--repo', '-p', type=str, default='.', help='Path to the Minecraft repo (defaults to the current working directory)')
	parser.add_argument('--message', '-m', type=str, default='version {}', help='Commit message template')
	parser.add_argument('--undo-renamed-vars', '-u', dest='undo_renamed_vars', action='store_true', help='Revert local variables that were renamed in the new version')
	parser.add_argument('version', nargs='+', type=str, help='List of mapping versions')
	parser.set_defaults(undo_renamed_vars=False)

	return parser.parse_args()


def undo_renames(repo: Repo) -> None:
	commit1 = repo.commit('HEAD')
	commit2 = None  # working tree
	diff_index = commit1.diff(commit2)

	# For each file changed (in reverse, with indices, so we can remove elements
	# in the for loop)
	for diff in diff_index:
		# Only process modified files (no new, deleted, ... files)
		if diff.change_type != 'M':
			continue

		# Only process Java files; leave everything else unchanged
		if not diff.a_path.endswith('.java'):
			continue

		source = get_blob(commit1, diff.a_path, repo_path)
		target = get_blob(commit2, diff.b_path, repo_path)

		try:
			renamed_variables = get_renamed_variables(source, target)
		except JavaAnalyzationError as e:
			raise Exception(f'{e} [{diff.a_path} -> {diff.b_path}]')

		if renamed_variables is not None:
			updated_target = undo_variable_renames(target, renamed_variables)
			with open(os.path.join(repo_path, diff.a_path), 'w') as f:
				f.write(updated_target)

			print(f'Updated {diff.a_path}')


def commit_version(repo: Repo, version: Version, undo_renamed_vars: bool, message_template: str) -> None:
	commit_msg = message_template.strip().format(str(version))
	if undo_renamed_vars and len(repo.iter_commits()) > 0:
		commit_msg += '\n\nRenamed variables reverted'

	repo.git.add('src')
	repo.git.commit('-m', commit_msg)


def create_version(repo: Repo, version: Version, undo_renamed_vars: bool, message_template: str) -> None:
	# 1. Generate source code for the current version
	print(f'Generating sources for Minecraft {version}')
	generate_sources(repo.working_tree_dir, version)

	# 2. If there are any previous versions, undo the renamed variables
	if undo_renamed_vars and len(repo.iter_commits()) > 0:
		print('Undoing renamed variables')
		undo_renames(repo)

	# 3. Commit the new version to git
	print('Committing to git')
	commit_version(repo, version, undo_renamed_vars, message_template)



import os
from shutil import move, rmtree
import subprocess

from git import Repo

from .version import Version


def generate_sources(source_repo: str, version: Version) -> None:
	script_dir = os.path.dirname(__file__)
	decompiler_dir = os.path.realpath(
		os.path.join(script_dir, '..', 'DecompilerMC')
	)

	dest_src_dir = os.path.join(source_repo, 'src')
	if os.path.exists(dest_src_dir):
		rmtree(dest_src_dir)

	# Generate source code
	try:
		for side in ('client', 'server'):
			p = subprocess.run(
				['python3', 'main.py', '--mcv', str(version), '-s', side, '-c', '-f', '-q'],
				stderr=subprocess.PIPE,
				cwd=decompiler_dir
			)
			if p.returncode != 0:
				raise Exception(p.stderr.decode())

		# Move the generated source code to the target repo
		move(
			os.path.join(decompiler_dir, 'src', str(version)),
			dest_src_dir
		)

	finally:
		# Remove large generated files so they won't end up in the build!
		for subdir in ('mappings', 'src', 'tmp', 'versions'):
			path = os.path.join(decompiler_dir, subdir)
			if os.path.exists(path):
				rmtree(path)



from typing import Dict

import pytest

from shulkr.minecraft.version import Version, load_manifest

MANIFEST_DATA = {
	'latest': {
		'release': '1.0.0',
		'snapshot': '1.0.0'
	},
	'versions': [
		{
			'type': 'release',
			'id': '1.0.0'
		},
		{
			'type': 'snapshot',
			'id': 'abcdef'
		}
	]
}

load_manifest(MANIFEST_DATA, earliest_supported_version_id='abcdef')


class TestVersion:
	snapshot = Version.of('abcdef')
	release = Version.of('1.0.0')

	def test_next_of_snapshot_is_the_release(self):
		assert self.snapshot.next == self.release

	def test_next_of_release_is_none(self):
		assert self.release.next is None

	def test_release_does_not_equal_snapshot(self):
		assert self.release != self.snapshot

	def test_snapshot_is_less_than_release(self):
		assert self.snapshot < self.release

	def test_release_is_greater_than_snapshot(self):
		assert self.release > self.snapshot

	def test_snapshot_to_release_returns_list_containing_snapshot_and_release(self):
		assert self.snapshot.to(self.release) == [self.snapshot, self.release]

	def test_snapshot_to_snapshot_returns_list_containing_snapshot(self):
		assert self.snapshot.to(self.snapshot) == [self.snapshot]

	def test_snapshot_to_none_returns_list_containing_snapshot_and_release(self):
		assert self.snapshot.to(None) == [self.snapshot, self.release]

	def test_snapshot_to_release_excluding_snapshots_returns_list_containing_release(self):
		assert self.snapshot.to(self.release, snapshots=False) == [self.release]

	def test_release_to_snapshot_raises_version_error(self):
		with pytest.raises(ValueError):
			self.release.to(self.snapshot)

	def test_pattern_with_no_range_operator_returns_list_containing_version(self):
		assert Version.pattern('1.0.0') == [self.release]

	def test_pattern_with_two_dots_returns_list_containing_release(self):
		assert Version.pattern('..') == [self.release]

	def test_pattern_with_three_dots_returns_list_containing_snapshot_and_release(self):
		assert Version.pattern('...') == [self.snapshot, self.release]

	def test_pattern_with_release_two_dots_returns_list_containing_release(self):
		assert Version.pattern('1.0.0..') == [self.release]

	def test_pattern_with_release_two_dots_release_returns_list_containing_release(self):
		assert Version.pattern('1.0.0..1.0.0') == [self.release]

	def test_pattern_with_snapshot_three_dots_release_returns_list_containing_snapshot_and_release(self):
		assert Version.pattern('abcdef...1.0.0') == [self.snapshot, self.release]



import shulkr

import git
import pytest


def test_create_version_with_undo_renamed_vars_on_repo_with_no_commits_does_not_call_undo_renames(mocker):
	# Mock
	mocker.patch('git.Repo')
	mocker.patch('shulkr.generate_sources')
	mocker.patch('shulkr.undo_renames')
	mocker.patch('shulkr.commit_version')

	repo = git.Repo()
	mocker.patch.object(repo, 'bare', return_value=False)

	# Call create_version
	shulkr.create_version(repo, '1.18.1', undo_renamed_vars=True, message_template=None)

	# Assert that undo_renames was not called
	shulkr.undo_renames.assert_not_called()


def test_create_version_with_undo_renamed_vars_on_repo_with_one_commit_calls_undo_renames(mocker):
	# Mock
	mocker.patch('git.Commit')
	mocker.patch('git.Repo')

	mocker.patch('shulkr.generate_sources')
	mocker.patch('shulkr.undo_renames')
	mocker.patch('shulkr.commit_version')

	repo = git.Repo()
	mocker.patch.object(repo, 'bare', return_value=False)
	mocker.patch.object(repo, 'iter_commits', return_value=[git.Commit()])

	# Call create_version
	shulkr.create_version(repo, '1.18.1', undo_renamed_vars=True, message_template=None)

	# Assert that undo_renames was not called
	shulkr.undo_renames.assert_called_once()
